use std::collections::HashMap;

use actix_web::{dev::PeerAddr, get, post, web, Error, HttpResponse};

use crate::{
    AppState, Claims, MqttClient, MqttDeviceDiscovery, MqttDeviceDiscoveryComponent,
    MqttDeviceDiscoveryDevice, MqttDeviceDiscoveryOrigin, MqttState, VERSION,
};

#[get("/config")]
pub async fn get_config(data: web::Data<AppState>, claims: Option<Claims>) -> HttpResponse {
    let claims = match claims {
        Some(x) => x,
        None => return HttpResponse::Unauthorized().finish(),
    };

    if !data.device_ids.contains(&claims.sub) {
        return HttpResponse::Unauthorized().finish();
    }

    let buf = match data.devices_api_json.get(&claims.sub) {
        Some(x) => x,
        None => return HttpResponse::InternalServerError().finish(),
    };

    HttpResponse::Ok()
        .content_type("application/json")
        .body(buf.clone())
}

#[post("/config")]
pub async fn post_config(
    claims: Option<Claims>,
    data: web::Data<AppState>,
    mqtt_client: web::Data<MqttClient>,
    device_state: web::Json<kdash_protocol::DeviceStatePost>,
) -> HttpResponse {
    let claims = match claims {
        Some(x) => x,
        None => return HttpResponse::Unauthorized().finish(),
    };

    if !data.device_ids.contains(&claims.sub) {
        return HttpResponse::Unauthorized().finish();
    }

    let (config, config_buf) = match (
        data.devices_api.get(&claims.sub),
        data.devices_api_json.get(&claims.sub),
    ) {
        (Some(config), Some(config_buf)) => (config, config_buf),
        _ => return HttpResponse::InternalServerError().finish(),
    };

    let device_id = claims.sub.to_string();

    let state_topic = format!("{}/device/{}/state", mqtt_client.topic, device_id);

    let state_payload = match serde_json::to_vec(&MqttState {
        battery_charging: device_state.battery_charging,
        battery_level: device_state.battery_level,
        battery_current: device_state.battery_current,
        battery_voltage: device_state.battery_voltage,
    }) {
        Ok(x) => x,
        Err(_) => return HttpResponse::InternalServerError().finish(),
    };

    if mqtt_client
        .client
        .publish(
            &state_topic,
            rumqttc::v5::mqttbytes::QoS::AtLeastOnce,
            false,
            state_payload,
        )
        .await
        .is_err()
    {
        return HttpResponse::InternalServerError().finish();
    }

    if let Some(discovery_topic) = &mqtt_client.discovery_topic {
        let unique_id = format!("kdash_{}", device_id);
        let unique_id_battery_level = format!("{}_battery_level", unique_id);
        let unique_id_battery_current = format!("{}_battery_current", unique_id);
        let unique_id_battery_voltage = format!("{}_battery_voltage", unique_id);

        let mut components = HashMap::new();
        components.insert(
            "battery_level".to_string(),
            MqttDeviceDiscoveryComponent {
                platform: "sensor",
                device_class: "battery",
                force_update: true,
                unit_of_measurement: Some("%"),
                value_template: "{{ value_json.battery_level }}",
                unique_id: &unique_id_battery_level,
            },
        );
        components.insert(
            "battery_current".to_string(),
            MqttDeviceDiscoveryComponent {
                platform: "sensor",
                device_class: "current",
                force_update: true,
                unit_of_measurement: Some("mA"),
                value_template: "{{ value_json.battery_current }}",
                unique_id: &unique_id_battery_current,
            },
        );
        components.insert(
            "battery_voltage".to_string(),
            MqttDeviceDiscoveryComponent {
                platform: "sensor",
                device_class: "voltage",
                force_update: true,
                unit_of_measurement: Some("mV"),
                value_template: "{{ value_json.battery_voltage }}",
                unique_id: &unique_id_battery_voltage,
            },
        );

        let discovery = MqttDeviceDiscovery {
            device: MqttDeviceDiscoveryDevice {
                identifiers: &unique_id,
                name: config.name.to_owned(),
            },
            origin: MqttDeviceDiscoveryOrigin {
                name: "kdash",
                software_version: VERSION,
                url: "https://git.dergrimm.net/dergrimm/kdash",
            },
            components,
            state_topic: &state_topic,
            qos: 2,
        };

        let discovery_payload = match serde_json::to_vec(&discovery) {
            Ok(x) => x,
            Err(_) => return HttpResponse::InternalServerError().finish(),
        };

        if mqtt_client
            .client
            .publish(
                format!("{}/device/kdash/{}/config", discovery_topic, device_id),
                rumqttc::v5::mqttbytes::QoS::AtLeastOnce,
                false,
                discovery_payload,
            )
            .await
            .is_err()
        {
            return HttpResponse::InternalServerError().finish();
        }
    }

    HttpResponse::Ok()
        .content_type("application/json")
        .body(config_buf.clone())
}

#[get("/image")]
pub async fn get_image(
    claims: Option<Claims>,
    peer_addr: Option<PeerAddr>,
    data: web::Data<AppState>,
    client: web::Data<reqwest::Client>,
) -> Result<HttpResponse, Error> {
    let claims = match claims {
        Some(x) => x,
        None => return Ok(HttpResponse::Unauthorized().finish()),
    };

    if !data.device_ids.contains(&claims.sub) {
        return Ok(HttpResponse::Unauthorized().finish());
    }

    let config = match data.devices.get(&claims.sub) {
        Some(x) => x,
        None => return Ok(HttpResponse::InternalServerError().finish()),
    };

    let mut forwarded_req = client.get(config.image.url.to_owned());
    forwarded_req = match peer_addr {
        Some(PeerAddr(addr)) => forwarded_req.header("x-forwarded-for", addr.ip().to_string()),
        None => forwarded_req,
    };

    let resp = forwarded_req
        .send()
        .await
        .map_err(actix_web::error::ErrorInternalServerError)?;

    let mut client_resp = HttpResponse::build(
        actix_web::http::StatusCode::from_u16(resp.status().as_u16())
            .map_err(actix_web::error::ErrorInternalServerError)?,
    );

    // Remove `Connection` as per
    // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Connection#Directives
    for (header_name, header_value) in resp.headers().iter().filter(|(h, _)| *h != "connection") {
        client_resp.insert_header((
            actix_web::http::header::HeaderName::from_bytes(header_name.as_ref()).unwrap(),
            actix_web::http::header::HeaderValue::from_bytes(header_value.as_ref()).unwrap(),
        ));
    }

    Ok(client_resp.streaming(resp.bytes_stream()))
}
